#include "libuvc/libuvc.h"
#include <stdio.h>

// really crappy tool to stream raw radiometric data,  based on libuvc's example.
// need patched libuvc.

struct cb_context {
  FILE *out;
  struct timeval tv_start;
  int frames;
};
 
void cb(uvc_frame_t *frame, void *ptr) {

  // just write out raw frames and display stats.
  uvc_error_t ret;
  struct cb_context *ctx = ptr;

  fwrite(frame->data, frame->data_bytes, 1, ctx->out);
  fflush(ctx->out);
  
  struct timeval tv;
  gettimeofday(&tv, 0);

  if (ctx->frames == 0) {
    ctx->tv_start = tv;
  }
  
  ctx->frames++;
  printf("got frame %d:%d, %d bytes\n", tv.tv_sec, tv.tv_usec, frame->data_bytes);
  
  
  if (ctx->frames == 100) {
    int elapsed = (tv.tv_sec - ctx->tv_start.tv_sec) * 1000000 + (tv.tv_usec - ctx->tv_start.tv_usec);
    if (elapsed != 0) {
        printf("%3.5f Hz (%d elapsed)\n", (1e6 / elapsed) * ctx->frames, elapsed);
    }
    ctx->frames = 0;
  }
}

int main(int argc, char **argv) {
  uvc_context_t *ctx;
  uvc_device_t *dev;
  uvc_device_handle_t *devh;
  uvc_stream_ctrl_t ctrl;
  uvc_error_t res;
  struct cb_context cb_ctx = {0};
  /* Initialize a UVC service context. Libuvc will set up its own libusb
   * context. Replace NULL with a libusb_context pointer to run libuvc
   * from an existing libusb context. */
  res = uvc_init(&ctx, NULL);
  if (res < 0) {
    uvc_perror(res, "uvc_init");
    return res;
  }
  puts("UVC initialized");
  
  if (argc < 2) {
    printf("usage: %s out.raw\n", *argv);
    return 1;
  }
  
  cb_ctx.out = fopen(argv[1], "wb");
  
  /* Locates the first attached UVC device, stores in dev */
  res = uvc_find_device(
      ctx, &dev,
      0, 0, NULL); /* filter devices: vendor_id, product_id, "serial_num" */
  if (res < 0) {
    uvc_perror(res, "uvc_find_device"); /* no devices found */
  } else {
    puts("Device found");
    /* Try to open the device: requires exclusive access */
    res = uvc_open(dev, &devh);
    if (res < 0) {
      uvc_perror(res, "uvc_open"); /* unable to open device */
    } else {
      puts("Device opened");
      /* Print out a message containing all the information that libuvc
       * knows about the device */
      uvc_print_diag(devh, stderr);
      /* Try to negotiate a 640x480 30 fps YUYV stream profile */
      res = uvc_get_stream_ctrl_format_size(
          devh, &ctrl, /* result stored in ctrl */
          UVC_FRAME_FORMAT_YUYV, /* YUV 422, aka YUV 4:2:2. try _COMPRESSED */
          320, 240, 15 /* width, height, fps */
      );
      /* Print out the result */
      uvc_print_stream_ctrl(&ctrl, stderr);
      if (res < 0) {
        uvc_perror(res, "get_mode"); /* device doesn't provide a matching stream */
      } else {
        /* Start the video stream. The library will call user function cb:
         *   cb(frame, (void*) 12345)
         */
        res = uvc_start_streaming(devh, &ctrl, cb, &cb_ctx, 0);
        if (res < 0) {
          uvc_perror(res, "start_streaming"); /* unable to start stream */
        } else {
          puts("Streaming...");
          uvc_set_ae_mode(devh, 1); /* e.g., turn on auto exposure */
          
          // stream until key is pressed.
          char c;
          read(0, &c, 1);
          /* End the stream. Blocks until last callback is serviced */
          uvc_stop_streaming(devh);
          puts("Done streaming.");
        }
      }
      /* Release our handle on the device */
      uvc_close(devh);
      puts("Device closed");
    }
    /* Release the device descriptor */
    uvc_unref_device(dev);
  }
  /* Close the UVC context. This closes and cleans up any existing device handles,
   * and it closes the libusb context if one was not provided. */
  uvc_exit(ctx);
  puts("UVC exited");
  return 0;
}
